# -*- coding: utf-8 -*-
# wasp_general/thread.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
from threading import Thread, Event
import traceback

from wasp_general.task.base import WStoppableTask, WTask
from wasp_general.verify import verify_type


class WThreadJoiningTimeoutError(Exception):
	""" Exception is raised when thread joining timeout is expired
	"""
	pass


class WThreadTask(WStoppableTask, metaclass=ABCMeta):
	""" Task that runs in a separate thread. Since there is no right way in Python to stop or terminate neighbor
	thread, so it's highly important for derived classes to be capable to stop correctly.

	This class implements :meth:`.WTask.start` method by creating new thread. Thread that is call
	:meth:`.WTask.thread_started` method.
	"""

	__thread_join_timeout__ = 10
	""" Default thread joining time (in seconds)
	"""

	__thread_name__ = None

	@verify_type(thread_name=(str, None), join_on_stop=bool, ready_to_stop=bool)
	@verify_type(thread_join_timeout=(int, float, None))
	def __init__(self, thread_name=None, join_on_stop=True, ready_to_stop=False, thread_join_timeout=None):
		""" Construct new threaded task.

		If 'ready_to_stop' is True, then thread event object is created (can be accessed through
		:meth:`.WThreadTask.ready_event`). This event shows, that this thread task has been done and it
		is ready to be stopped correctly. This event is set automatically after :meth:`.WTask.thread_started`
		method call. So, it implies that this method doesn't call extra thread or process creation,
		or :meth:`.WTask.thread_started` method waits for a thread/process termination. It implies that there
		are no leftover threads or processes (which can be cleaned up later, in the
		:meth:`.WTask.thread_stopped` method).

		If 'join_on_stop' is True, then thread event object is created (can be accessed through
		:meth:`.WThreadTask.stop_event`). This event shows, that there was a request for task to stop. Since
		task can be requested to stop at any time (application terminated, task canceled, ...) , it is
		better to enable and poll this flag. This flag enables automatic call of thread cleanup. When
		this flag is False, then child class must do all the cleaning itself (like
		thread joining and :meth:`.WThreadTask.close_thread` method calling).

		There are two other events - :meth:`.WThreadTask.start_event` and :meth:`.WThreadTask.exception_event`.
		The first one shows that thread function was started (it means that new thread was already created).
		The second one is set when exception is raised inside :meth:`.WThreadTask.thread_started` method.
		All exceptions, that are raised inside thread function, are passed to the
		callback :meth:`.WThreadTask.thread_exception`.

		With both flags ('ready_to_stop' and 'join_on_stop') there can be a situation, when ready event
		wasn't set, but stop event has been set already. This situation shows, that task was terminated
		before completion.

		:note: With join_on_stop flag enabled, :meth:`.WThreadTask.stop` method can not be called from the same
		execution thread. It means, that it can not be called from :meth:`.WThreadTask.start` or
		:meth:`.WThreadTask.thread_started` methods in direct or indirect way. In that case it is better to use
		'ready_to_stop' event polling.

		:param thread_name: name of the thread. It is used in thread constructor as name value only
		:param join_on_stop: define whether to create stop event object or not.
		:param ready_to_stop: define whether to create ready event object or not
		:param thread_join_timeout: timeout for joining operation. If it isn't set then default \
		:attr:`.WThreadTask.__thread_join_timeout__` value will be used. This value is used in \
		:meth:`.WThreadTask.close_thread` method and if thread wasn't stopped for this period of time, then \
		:class:`.WThreadJoiningTimeoutError` exception will be raised.

		note: Most event objects are cleared at :meth:`.WThreadTask.start` method (such as 'ready_event',
		'stop_event', 'exception_event'). But only 'start_event' is cleared at stopping process. It is made
		this way because there may be concurrency if multiple threads that will wait for this thread to
		stop and one of those threads will clear the flag/event before other threads will do their job.
		"""
		WStoppableTask.__init__(self)

		self.__thread_join_timeout = self.__class__.__thread_join_timeout__
		if thread_join_timeout is not None:
			self.__thread_join_timeout = thread_join_timeout
		self.__thread = None
		self.__thread_name = thread_name if thread_name is not None else self.__class__.__thread_name__
		self.__start_event = Event()
		self.__stop_event = Event() if join_on_stop is True else None
		self.__ready_event = Event() if ready_to_stop is True else None
		self.__exception_event = Event()

	def thread(self):
		""" Return current Thread object (or None if task wasn't started)

		:return: Thread or None
		"""
		return self.__thread

	def thread_name(self):
		""" Return thread name with which this thread is or will be created

		:return: str
		"""
		return self.__thread_name

	def start_event(self):
		""" Return event which is set after the thread creation. Shows that a separate thread has been created
		already

		:return: Event
		"""
		return self.__start_event

	def stop_event(self):
		""" Return stop event object. Event will be available if object was constructed with join_on_stop flag

		:return: Event or None
		"""
		return self.__stop_event

	def ready_event(self):
		""" Return readiness event object. Event will be available if object was constructed with ready_to_stop
		flag

		:return: Event or None
		"""
		return self.__ready_event

	def exception_event(self):
		""" Return event which is set if exception is raised inside thread function

		:return: Event
		"""
		return self.__exception_event

	def join_timeout(self):
		""" Return task join timeout

		:return: int or float
		"""
		return self.__thread_join_timeout

	def close_thread(self):
		""" Clear all object descriptors for stopped task. Task must be joined prior to calling this method.

		:return: None
		"""
		if self.__thread is not None and self.__thread.is_alive() is True:
			raise WThreadJoiningTimeoutError('Thread is still alive. Thread name: %s' % self.__thread.name)
		self.start_event().clear()
		self.__thread = None

	def start(self):
		""" :meth:`WStoppableTask.start` implementation that creates new thread
		"""

		start_event = self.start_event()
		stop_event = self.stop_event()
		ready_event = self.ready_event()

		def thread_target():
			try:
				start_event.set()
				self.thread_started()
				if ready_event is not None:
					ready_event.set()
			except Exception as e:
				self.exception_event().set()
				self.thread_exception(e)

		if self.__thread is None:
			if stop_event is not None:
				stop_event.clear()

			if ready_event is not None:
				ready_event.clear()

			self.exception_event().clear()

			self.__thread = Thread(target=thread_target, name=self.thread_name())
			self.__thread.start()

	def stop(self):
		""" :meth:`WStoppableTask.stop` implementation that sets stop even (if available), calls
		:meth:`WStoppableTask.threaded_stopped` and cleans up thread (if configured)
		"""
		thread = self.thread()
		if thread is not None:
			if self.stop_event() is not None:
				self.stop_event().set()

			self.thread_stopped()

			if self.stop_event() is not None:
				thread.join(self.join_timeout())
				self.close_thread()

	@abstractmethod
	def thread_started(self):
		""" Real task that do all the work
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def thread_stopped(self):
		""" Method is called when task is about to stop (is called before joining process).
		This method is called whenever exception was raised or not

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	def thread_exception(self, raised_exception):
		""" Callback for handling exception, that are raised inside :meth:`.WThreadTask.thread_started`

		:param raised_exception: raised exception
		:return: None
		"""
		print('Thread execution was stopped by the exception. Exception: %s' % str(raised_exception))
		print('Traceback:')
		print(traceback.format_exc())

	def check_events(self):
		""" Check "stopping"-events ('ready_event', 'stop_event', 'exception_event') if one of them is set.
		Usually True value means that thread is meant to be stopped, means that it is finished its job or
		some error has happened or this thread was asked to stop

		:return: bool
		"""
		return (
			self.ready_event().is_set() is True or
			self.stop_event().is_set() is True or
			self.exception_event().is_set() is True
		)


class WThreadCustomTask(WThreadTask):
	""" Class that can run any task in a separate thread. It just wraps start method, and for a
	:class:`.WStoppableTask` object it wraps stop method also. So for a WThreadTask class task, this object
	will create new thread "inside" new thread. Because of this, it is important that appropriate flags was set
	within constructor
	"""

	@verify_type(task=WTask)
	@verify_type('paranoid', thread_name=(str, None), join_on_stop=bool, ready_to_stop=bool)
	@verify_type('paranoid', thread_join_timeout=(int, float, None))
	def __init__(self, task, thread_name=None, join_on_stop=True, ready_to_stop=False, thread_join_timeout=None):
		""" Create new WThreadTask task for the given task

		:param task: task that must be started in a separate thread
		:param thread_name: same as thread_name in :meth:`.WThreadTask.__init__` method
		:param join_on_stop: same as join_on_stop in :meth:`.WThreadTask.__init__` method
		:param ready_to_stop: same as ready_to_stop in :meth:`.WThreadTask.__init__` method
		:param thread_join_timeout: same as thread_join_timeout in :meth:`.WThreadTask.__init__` method
		"""
		WThreadTask.__init__(
			self, thread_name=thread_name, join_on_stop=join_on_stop, ready_to_stop=ready_to_stop,
			thread_join_timeout=thread_join_timeout
		)
		self.__task = task

	def task(self):
		""" Return original task

		:return: WTask
		"""
		return self.__task

	def thread_started(self):
		""" Start original task

		:return: None
		"""
		self.task().start()

	def thread_stopped(self):
		""" If original task is :class:`.WStoppableTask` object, then stop it

		:return: None
		"""
		task = self.task()
		if isinstance(task, WStoppableTask) is True:
			task.stop()


class WPollingThreadTask(WThreadTask, metaclass=ABCMeta):
	""" Create task, that will be executed in a separate thread, and will wait for stop event or ready event and
	till that will do small piece of work. This threaded task will be constructed with
	'join_on_stop' and 'ready_to_stop' flags turned on

	Polling timeout is a timeout after which next call for a small piece of work will be done. Real
	:meth:`.WPollingThreadTask.__polling_iteration` method implementation must be fast
	(faster then joining timeout), so it must do small piece of work each time only. It is crucial to do that,
	because busy thread can be terminated at any time, and so can not be finalized gracefully.

	If one thread spawns other threads it is obvious to stop them from the same thread they are generated.
	And at this point, wrong joining and polling timeouts could break start-stop mechanics. So parent thread
	should have joining timeout not less then children threads have (it is better to have joining timeout greater
	then children timeout). And polling timeout should be not greater (as little as possible) then children
	threads have
	"""

	__thread_polling_timeout__ = WThreadTask.__thread_join_timeout__ / 4
	""" Default polling timeout
	"""

	@verify_type('paranoid', thread_name=(str, None), thread_join_timeout=(int, float, None))
	@verify_type(polling_timeout=(int, float, None))
	def __init__(self, thread_name=None, thread_join_timeout=None, polling_timeout=None):
		""" Create new task

		:param thread_name: same as 'thread_name' in :meth:`.WThreadTask.__init__`
		:param thread_join_timeout: same as 'thread_join_timeout' in :meth:`.WThreadTask.__init__`
		:param polling_timeout: polling timeout for this task
		"""
		WThreadTask.__init__(
			self, thread_name=thread_name, join_on_stop=True, ready_to_stop=True,
			thread_join_timeout=thread_join_timeout
		)
		self.__polling_timeout = \
			polling_timeout if polling_timeout is not None else self.__class__.__thread_polling_timeout__

	def polling_timeout(self):
		""" Task polling timeout

		:return: int or float
		"""
		return self.__polling_timeout

	def thread_started(self):
		""" Start polling for a stop event or ready event and do small work via
		:meth:`.WPollingThreadTask._polling_iteration` method call

		:return: None
		"""
		while self.check_events() is False:
			self._polling_iteration()
			self.stop_event().wait(self.polling_timeout())

	@abstractmethod
	def _polling_iteration(self):
		""" Do small work

		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WThreadedTaskChain(WPollingThreadTask):
	""" Threaded task, that executes given tasks sequentially
	"""

	@verify_type(threaded_task_chain=WThreadTask)
	@verify_type('paranoid', thread_name=(str, None), thread_join_timeout=(int, float, None))
	@verify_type('paranoid', polling_timeout=(int, float, None))
	def __init__(
		self, *threaded_task_chain, thread_name=None, thread_join_timeout=None, polling_timeout=None
	):
		""" Create threaded tasks

		:param threaded_task_chain: tasks to execute
		:param thread_name: same as thread_name in :meth:`WPollingThreadTask.__init__`
		:param thread_join_timeout: same as thread_join_timeout in :meth:`WPollingThreadTask.__init__`
		:param polling_timeout: same as polling_timeout in :meth:`WPollingThreadTask.__init__`
		"""
		WPollingThreadTask.__init__(
			self, thread_name=thread_name, thread_join_timeout=thread_join_timeout,
			polling_timeout=polling_timeout
		)
		self.__task_chain = threaded_task_chain
		for task in self.__task_chain:
			if task.ready_event() is None:
				raise ValueError("Chained task must be constructed with 'ready_to_stop' flag")
		self.__current_task = None

	def _polling_iteration(self):
		""" :meth:`.WPollingThreadTask._polling_iteration` implementation
		"""
		if len(self.__task_chain) > 0:
			if self.__current_task is None:
				self.__current_task = 0

			task = self.__task_chain[self.__current_task]
			if task.thread() is None:
				task.start()
			elif task.ready_event().is_set() is True:
				task.stop()
				if self.__current_task < (len(self.__task_chain) - 1):
					self.__current_task += 1
				else:
					self.ready_event().set()
			elif task.exception_event().is_set() is True:
				raise RuntimeError('Child thread failed')
		else:
			self.ready_event().set()

	def thread_stopped(self):
		""" :meth:`.WThreadTask._polling_iteration` implementation
		"""
		if self.__current_task is not None:
			task = self.__task_chain[self.__current_task]
			task.stop()
			self.__current_task = None



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/layers.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import abstractmethod
from enum import Enum

from wasp_general.verify import verify_type, verify_subclass

from wasp_general.network.messenger.proto import WMessengerEnvelopeProto
from wasp_general.network.messenger.proto import WMessengerOnionSessionProto, WMessengerOnionLayerProto

from wasp_general.network.messenger.envelope import WMessengerTextEnvelope, WMessengerBytesEnvelope


class WMessengerSimpleCastingLayer(WMessengerOnionLayerProto):

	__layer_name__ = "com.binblob.wasp-general.simple-casting-layer"
	""" Layer name
	"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionLayerProto.__init__(self, WMessengerSimpleCastingLayer.__layer_name__)

	@verify_type(envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	@verify_subclass(from_envelope=WMessengerEnvelopeProto, to_envelope=WMessengerEnvelopeProto)
	def process(self, envelope, session, from_envelope=None, to_envelope=None, **kwargs):
		if isinstance(envelope, from_envelope) is False:
			raise TypeError('Source envelope type mismatch')
		return to_envelope(envelope.message(), meta=envelope)


class WMessengerOnionModeLayerProto(WMessengerOnionLayerProto):
	""" Simple WMessengerOnionLayerProto implementation, that can have different message processing mechanisms
	(depends on the "mode" value). This "mode" must be always specified as
	mode argument in :meth:`.WMessengerOnionModeLayerProto.process` method. This argument must be the same type
	(or be a subclass of the type), that is specified in constructor
	"""

	@verify_type(name=str, mode_cls=type)
	def __init__(self, name, mode_cls):
		""" Construct new layer

		:param name: layer name
		:param mode_cls: layer's "mode" class
		"""
		WMessengerOnionLayerProto.__init__(self, name)
		self.__mode_cls = mode_cls

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	def process(self, envelope, session, mode=None, **kwargs):
		""" :meth:`.WMessengerOnionLayerProto.process` implementation
		"""
		if mode is None:
			raise RuntimeError('"mode" argument must be specified for this object')

		if isinstance(mode, self.__mode_cls) is False:
			raise TypeError('Invalid "mode" argument')

		return self._process(envelope, session, mode, **kwargs)

	@abstractmethod
	@verify_type(envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	def _process(self, envelope, session, mode, **kwargs):
		""" Real processing method.

		:param envelope: original envelope
		:param session:  original session
		:param mode: specified mode
		:param kwargs: layer arguments

		:return: WMessengerEnvelopeProto
		"""
		raise NotImplementedError('This method is abstract')


class WMessengerOnionCoderLayerProto(WMessengerOnionModeLayerProto):
	""" Class for layers, that are used for encryption/decryption, encoding/decoding. This layer class works with
	strings and bytes and as a result generates strings and bytes
	"""

	class Mode(Enum):
		""" Specifies layers mode
		"""
		encode = 1
		""" Encryption/encoding mode
		"""
		decode = 2
		""" Decryption/decoding mode
		"""

	@verify_type(name=str)
	def __init__(self, name):
		""" Construct new layer

		:param name: layer name
		"""
		WMessengerOnionModeLayerProto.__init__(self, name, WMessengerOnionCoderLayerProto.Mode)

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	def _process(self, envelope, session, mode, **kwargs):
		""" :meth:`.WMessengerOnionLayerProto.process` implementation
		"""
		if mode == WMessengerOnionCoderLayerProto.Mode.encode:
			return self.encode(envelope, session, **kwargs)
		else:  # mode == WMessengerOnionCoderLayerProto.Mode.decode
			return self.decode(envelope, session, **kwargs)

	@abstractmethod
	@verify_type(envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope), session=WMessengerOnionSessionProto)
	def encode(self, envelope, session, **kwargs):
		""" Encrypt/encode message

		:param envelope: message to encrypt/encode
		:param session: original session
		:return: WMessengerTextEnvelope or WMessengerBytesEnvelope
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope), session=WMessengerOnionSessionProto)
	def decode(self, envelope, session, **kwargs):
		""" Decrypt/decode message

		:param envelope: message to decrypt/decode
		:param session: original session
		:return: WMessengerTextEnvelope or WMessengerBytesEnvelope
		"""
		raise NotImplementedError('This method is abstract')


class WMessengerOnionPackerLayerProto(WMessengerOnionModeLayerProto):
	""" Class for layers, that are used for packing/unpacking, serializing/de-serializing. This layer class
	can pack "any" envelope and produce WMessengerTextEnvelope or WMessengerBytesEnvelope or can
	unpack WMessengerTextEnvelope (or WMessengerBytesEnvelope) to "any" object

	(not "any" but the most)
	"""

	class Mode(Enum):
		""" Specifies layers mode
		"""
		pack = 1
		""" Encryption/encoding mode
		"""
		unpack = 2
		""" Decryption/decoding mode
		"""

	@verify_type(name=str)
	def __init__(self, name):
		WMessengerOnionModeLayerProto.__init__(self, name, WMessengerOnionPackerLayerProto.Mode)

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	def _process(self, envelope, session, mode, **kwargs):
		""" :meth:`.WMessengerOnionLayerProto.process` implementation
		"""
		if mode == WMessengerOnionPackerLayerProto.Mode.pack:
			return self.pack(envelope, session, **kwargs)
		else:  # mode == WMessengerOnionPackerLayerProto.Mode.unpack
			return self.unpack(envelope, session, **kwargs)

	@abstractmethod
	@verify_type(envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	def pack(self, envelope, session, **kwargs):
		""" Pack/serialize message

		:param envelope: message to pack/serialize
		:param session: original session
		:return: WMessengerTextEnvelope or WMessengerBytesEnvelope
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope), session=WMessengerOnionSessionProto)
	def unpack(self, envelope, session, **kwargs):
		""" Unpack/de-serialize message

		:param envelope: message to unpack/de-serialize
		:param session: original session
		:return: WMessengerEnvelopeProto
		"""
		raise NotImplementedError('This method is abstract')





# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.random import random_bits, random_int, random_bytes


def test_random_bits():
	assert(isinstance(random_bits(3), int) is True)
	assert(random_bits(0) == 0)
	assert(random_bits(10) != random_bits(10))


def test_random_int():
	assert(isinstance(random_int(10), int) is True)
	assert(random_int(0) == 0)
	assert(random_int(1000) != random_int(1000))
	assert(random_int(10) <= 10)


def test_random_bytes():
	pytest.raises(TypeError, random_bytes)
	pytest.raises(TypeError, random_bytes, '1')
	pytest.raises(ValueError, random_bytes, -1)

	assert(isinstance(random_bytes(0), bytes) is True)
	assert(isinstance(random_bytes(3), bytes) is True)
	assert(random_bytes(0) == b'')
	assert(len(random_bytes(10)) == 10)
	assert(random_bytes(10) != random_bytes(10))



# -*- coding: utf-8 -*-

import pytest
import io

from wasp_general.crypto.aes import WBlockPadding, WSimplePadding, WZeroPadding, WShiftPadding, WPKCS7Padding, WAESMode
from wasp_general.crypto.aes import WAES


def test_abstract_classes():
	pytest.raises(TypeError, WBlockPadding)
	pytest.raises(NotImplementedError, WBlockPadding.pad, None, b'', 1)
	pytest.raises(NotImplementedError, WBlockPadding.reverse_pad, None, b'', 1)


class TestWSimplePadding:

	def test_padding(self):
		assert(issubclass(WSimplePadding, WBlockPadding) is True)
		assert(WSimplePadding().padding_symbol() == b'\x00')
		assert(WSimplePadding(23).padding_symbol() == b'\x17')

		padding = WSimplePadding(23)
		assert(padding.pad(b'123', 3) == b'123')
		assert(padding.pad(b'123', 7) == b'123\x17\x17\x17\x17')
		assert(padding.reverse_pad((b'123' + (chr(23) * 4).encode()), 7) == b'123')


class TestWZeroPadding:

	def test(self):
		padding = WZeroPadding()
		assert(isinstance(padding, WZeroPadding) is True)
		assert(isinstance(padding, WSimplePadding) is True)
		assert(padding.padding_symbol() == b'\x00')


class TestWShiftPadding:

	def test_padding(self):
		assert(issubclass(WShiftPadding, WSimplePadding) is True)

		padding = WShiftPadding(21)
		assert(padding.pad(b'123', 4) in [b'\x15123', b'123\x15'])
		assert(padding.reverse_pad(b'\x15123', 4) == b'123')
		assert(padding.reverse_pad(b'123\x15', 4) == b'123')


class TestWPKCS7Padding:

	def test_padding(self):
		assert(issubclass(WPKCS7Padding, WBlockPadding) is True)

		padding = WPKCS7Padding()
		assert(padding.pad(b'123', 6) == b'123\x03\x03\x03')

		assert(padding.reverse_pad(b'123\x02\x02', 5) == b'123')
		assert(padding.reverse_pad(b'123\x03\x03\x03', 6) == b'123')

		pytest.raises(ValueError, padding.reverse_pad, b'123\x10', 4)
		pytest.raises(ValueError, padding.reverse_pad, b'123\x02', 4)


class TestWAESMode:

	def test_mode(self):

		pytest.raises(ValueError, WAESMode, 16, 'AES-CBC', b'', padding=WPKCS7Padding())

		assert(WAESMode.init_sequence_length(16, 'AES-CBC') == 32)
		assert(WAESMode.init_sequence_length(24, 'AES-CTR') == 40)

		assert(WAESMode.parse_cipher_name('aes-256-cbc') == (32, 'AES-CBC'))
		assert(WAESMode.parse_cipher_name('AES_128_CTR') == (16, 'AES-CTR'))
		pytest.raises(ValueError, WAESMode.parse_cipher_name, '????')
		pytest.raises(ValueError, WAESMode.parse_cipher_name, 'AES-32-cbc')
		pytest.raises(ValueError, WAESMode.parse_cipher_name, 'AES-128-cbc2')

		init_seq_key = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
		init_seq_iv = b'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
		padding = WPKCS7Padding()

		cbc_mode = WAESMode(16, 'AES-CBC', init_seq_key + init_seq_iv, padding=padding)
		assert(cbc_mode.key_size() == 16)
		assert(cbc_mode.mode() == 'AES-CBC')
		assert(cbc_mode.padding() == padding)
		assert(cbc_mode.initialization_vector() == init_seq_iv)

		assert(isinstance(cbc_mode.pyaes_args(), tuple) is True)
		assert(isinstance(cbc_mode.pyaes_kwargs(), dict) is True)

		init_seq_key = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
		init_seq_key += b'\x10\x11\x12\x13\x14\x15\x16\x17'
		init_seq_counter = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07'

		ctr_mode = WAESMode(24, 'AES-CTR', init_seq_key + init_seq_counter)
		assert(ctr_mode.key_size() == 24)
		assert(ctr_mode.mode() == 'AES-CTR')
		assert(ctr_mode.padding() is None)
		assert (ctr_mode.initialization_counter_value() == 7)


@pytest.fixture
def fake_aes_mode(request):
	modes = WAESMode.__modes_descriptor__.copy()
	WAESMode.__modes_descriptor__['TEST-AES-MODE'] = {
		'mode_code': None,
		'requirements': {
			'initialization_vector': True,
			'counter': True
		}
	}

	def fin():
		WAESMode.__modes_descriptor__ = modes
	request.addfinalizer(fin)


class TestWAES:

	def test_cipher(self):

		init_seq_key1 = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
		init_seq_iv = b'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
		padding = WPKCS7Padding()

		cbc_mode1 = WAESMode(16, 'AES-CBC', init_seq_key1 + init_seq_iv, padding=padding)

		init_seq_key2 = b'\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f'
		cbc_mode2 = WAESMode(16, 'AES-CBC', init_seq_key2 + init_seq_iv, padding=padding)

		a1 = WAES(cbc_mode1)
		a2 = WAES(cbc_mode2)

		text_block = 'q' * a1.mode().key_size()

		c1 = a1.cipher()
		c2 = a2.cipher()
		assert(c1.encrypt(text_block) != text_block.encode())

		c1 = a1.cipher()
		c2 = a2.cipher()
		assert(c1.encrypt(text_block) != c2.encrypt(text_block))

		c1 = a1.cipher()
		c2 = a1.cipher()
		assert(c1.decrypt(c1.encrypt(text_block)) != text_block.encode())

		c1 = a1.cipher()
		c2 = a1.cipher()
		assert(c2.decrypt(c1.encrypt(text_block)) == text_block.encode())

		c1 = a1.cipher()
		c2 = a2.cipher()
		assert(c2.decrypt(c1.encrypt(text_block)) != text_block.encode())

		text_block = b'qwerty'
		assert(a1.encrypt(text_block) == a1.encrypt(text_block))
		assert(a1.encrypt(text_block) != text_block)
		assert(a1.decrypt(a1.encrypt(text_block)) == text_block)
		assert(a1.decrypt(a1.encrypt(text_block), decode=True) == text_block.decode())

	@pytest.mark.usefixtures('fake_aes_mode')
	def test_fake_mode(self):
		init_seq_key = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
		init_seq_iv = b'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
		init_seq_counter = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08'

		fake_mode = WAESMode(16, 'TEST-AES-MODE', init_seq_key + init_seq_iv + init_seq_counter)
		assert(fake_mode.initialization_vector() == init_seq_iv)
		assert(fake_mode.initialization_counter_value() == 8)
